/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2005, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

Count the number of sequential like records in a file, and print the
count to stdout; the records in the file must be in lexical order, as
per sort(1).

The output on stdout is the count of the number of sequential records,
followed by a tab, followed by the record.

$Revision: 0.0 $
$Date: 2005/09/12 21:44:42 $
$Id: tscount.c,v 0.0 2005/09/12 21:44:42 john Exp $
$Log: tscount.c,v $
Revision 0.0  2005/09/12 21:44:42  john
Initial revision


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static char rcsid[] = "$Id: tscount.c,v 0.0 2005/09/12 21:44:42 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2005, John Conover, All Rights Reserved"; /* the copyright banner */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Count the number of like records in a file\n",
    "Usage: tscount [-r] [-v] [filename]\n",
    "    -r, reverse the output fields\n",
    "    -v, print the program's version information\n",
    "    filename, input filename\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening file\n",
    "Error closing file\n",
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define EOPEN 2
#define ECLOSE 3

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */

#else

static void print_message (); /* print any error messages */

#endif

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char buffer[BUFSIZ + 1], /* input buffer */
         last_buffer[BUFSIZ + 1], /* last input buffer */
         *eol_ref; /* reference to '\n' in buffer */

    int count = 0, /* running count of like records */
        retval = NOERROR, /* return value, assume no error */
        reverse = 0, /* reverse the output fields flag, 0 = no, 1 = yes */
        c; /* command line switch */

    FILE *infile = stdin; /* reference to input file */

    while ((c = getopt (argc, argv, "rv")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'r': /* request for reverse the output fields? */

                reverse = 1; /* yes, set the reverse the output fields flag, 0 = no, 1 = yes */
                break;

            case 'v': /* request for version information? */

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* enough arguments? */
    {
        retval = EOPEN; /* assume error opening file */

        if ((infile = argc <= optind ? stdin : fopen (argv[optind], "r")) != (FILE *) 0) /* yes, open the input file */
        {
            retval = NOERROR; /* assume no error */

            if (fgets (buffer, BUFSIZ, infile) != (char *) 0) /* get the next record from the file */
            {

                if ((eol_ref = strchr (buffer, '\n')) != (char *) 0) /* find the EOL character */
                {
                    *eol_ref = '\0'; /* null the EOL character */
                }

                else
                {
                    (void) fprintf (stderr, "Buffer overflow\n");
                    exit (1);
                }

                (void) strcpy (last_buffer, buffer); /* save the input record */
                count ++; /* one more like record */

                while (fgets (buffer, BUFSIZ, infile) != (char *) 0)
                {

                    if ((eol_ref = strchr (buffer, '\n')) != (char *) 0) /* find the EOL character */
                    {
                        *eol_ref = '\0'; /* null the EOL character */
                    }

                    else
                    {
                        (void) fprintf (stderr, "Buffer overflow\n");
                        exit (1);
                    }

                    if (strcmp (last_buffer, buffer) != 0) /* input record like the last input record? */
                    {

                        if (reverse == 0) /* reverse the output fields flag, 0 = no, 1 = yes, set? */
                        {
                            (void) printf ("%d\t%s\n", count, last_buffer); /* print the count and its record */
                        }

                        else
                        {
                            (void) printf ("%s\t%d\n", last_buffer, count); /* print the record and its count */
                        }

                        count = 0; /* reset the running count of like records */
                    }

                    (void) strcpy (last_buffer, buffer); /* save the input record */
                    count ++; /* one more like record */
                }

                if (reverse == 0) /* reverse the output fields flag, 0 = no, 1 = yes, set? */
                {
                    (void) printf ("%d\t%s\n", count, last_buffer); /* print the count and its record */
                }

                else
                {
                    (void) printf ("%s\t%d\n", last_buffer, count); /* print the record and its count */
                }

            }

            if (argc > optind) /* using stdin as input? */
            {

                if (fclose (infile) == EOF) /* no, close the input file */
                {
                    retval = ECLOSE; /* error closing file */
                }

            }

        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}
